theory Embedding_Trs
imports
  Trs
  "$AFP/Well_Quasi_Orders/Kruskal"
begin

inductive_set gterms for A where
  "f \<in> A \<Longrightarrow> \<forall>t\<in>set ts. t \<in> gterms A \<Longrightarrow> Fun f ts \<in> gterms A"

text {*Ground-terms are finite trees.*}
interpretation term_tree: finite_tree Fun "fst \<circ> the \<circ> root" args
  by (unfold_locales) simp_all

text {*The set @{term "finite_tree.trees Fun A"} is the same
as the set of ground-terms over @{term A}.*}
lemma gterms_simp [simp]:
  "gterms A = term_tree.trees A"
  (is "?lhs = ?rhs")
proof
  show "?lhs \<subseteq> ?rhs"
  proof
    fix t assume "t \<in> ?lhs" thus "t \<in> ?rhs"
      by (induct) (rule term_tree.trees_intros, auto)
  qed
next
  show "?rhs \<subseteq> ?lhs"
  proof
    fix t assume "t \<in> ?rhs" thus "t \<in> ?lhs"
      by (induct) (auto intro: gterms.intros)
  qed
qed

definition Emb where
  "Emb \<equiv> { (Fun f ts, t) | f ts t. t \<in> set ts }"

definition Hemb where
  "Hemb P \<equiv> Emb \<union>
    { (Fun f ss, Fun g ts) | f ss g ts. P\<^sup>=\<^sup>= g f \<and> sublisteq ts ss }"

lemma list_hembeq_mono:
  assumes "\<And>s t. P s t \<longrightarrow> Q s t"
  shows "list_hembeq P s t \<longrightarrow> list_hembeq Q s t"
proof
  assume "list_hembeq P s t" thus "list_hembeq Q s t"
    by (induct) (insert assms, auto)
qed

lemma reflclp_mono:
  assumes "\<And>s t. P s t \<longrightarrow> Q s t"
  shows "P\<^sup>=\<^sup>= s t \<longrightarrow> Q\<^sup>=\<^sup>= s t"
  using assms by auto

abbreviation hemb where "hemb P \<equiv> term_tree.tree_hembeq P"

text {*Rewriting step without substitution.*}
inductive_set cstep for R :: "('a, 'b) trs" where
  base [intro]: "(l, r) \<in> R \<Longrightarrow> (l, r) \<in> cstep R" |
  step [intro]: "(s, t) \<in> cstep R \<Longrightarrow> (C\<langle>s\<rangle>, C \<langle>t\<rangle>) \<in> cstep R"

lemma cstepD:
  assumes "(s, t) \<in> cstep R"
  shows "\<exists>l r C. (l, r) \<in> R \<and> s = C\<langle>l\<rangle> \<and> t = C\<langle>r\<rangle>"
using assms
proof (induct)
  case (base l r)
  hence "(l, r) \<in> R \<and> l = \<box>\<langle>l\<rangle> \<and> r = \<box>\<langle>r\<rangle>" by simp
  thus ?case by blast
next
  case (step s t C)
  then obtain l r D where "(l, r) \<in> R"
    and s: "s = D\<langle>l\<rangle>" and t: "t = D \<langle>r\<rangle>" by auto
  moreover have "C\<langle>s\<rangle> = (C \<circ>c D)\<langle>l\<rangle>" and "C\<langle>t\<rangle> = (C \<circ>c D)\<langle>r\<rangle>" by (simp_all add: s t)
  ultimately show ?case by blast
qed

lemma cstepE [elim]:
  assumes "(s, t) \<in> cstep R"
    and "\<And>C l r. (l, r) \<in> R \<Longrightarrow> s = C\<langle>l\<rangle> \<Longrightarrow> t = C\<langle>r\<rangle> \<Longrightarrow> P"
  shows P
  using assms and cstepD by blast

lemma hemb_ctxt:
  assumes "hemb P s t"
  shows "hemb P (C\<langle>s\<rangle>) (C\<langle>t\<rangle>)"
  using assms by (induct C) auto

lemma list_hembeq_cstep_HembD:
  assumes "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb P))\<^sup>*) xs zs"
    (is "list_hembeq (\<lambda>s t. (t, s) \<in> ?R) xs zs" is "list_hembeq ?P xs zs")
  shows "\<exists>ys. sublisteq ys zs \<and> length ys = length xs \<and>
    (\<forall>i<length ys. (ys ! i, xs ! i) \<in> (cstep (Hemb P))\<^sup>*)"
using assms
proof (induct)
  case (list_hembeq_Nil ys)
  thus ?case by auto
next
  case (list_hembeq_Cons xs zs z)
  then obtain ys where list_hembeq: "sublisteq ys zs" and "length ys = length xs"
    and "\<forall>i<length ys. (ys ! i, xs ! i) \<in> ?R" by auto
  moreover have "sublisteq ys (z # zs)" using list_hembeq by auto
  ultimately show ?case by blast
next
  case (list_hembeq_Cons2 x z xs zs)
  then obtain ys where list_hembeq: "sublisteq ys zs"
    and len: "length ys = length xs"
    and *: "\<forall>i<length ys. (ys ! i, xs ! i) \<in> ?R" by auto
  let ?ys = "z # ys" and ?xs = "x # xs"
  from list_hembeq have "sublisteq ?ys (z # zs)" by auto
  moreover have "length ?ys = length ?xs" using len by auto
  moreover have "\<forall>i<length ?ys. (?ys ! i, ?xs ! i) \<in> ?R"
  proof (intro allI impI)
    fix i
    assume i: "i < length ?ys"
    show "(?ys ! i, ?xs ! i) \<in> ?R"
      using i and * and `(\<lambda>s t. (t, s) \<in> ?R)\<^sup>=\<^sup>= x z`
      by (cases i) auto
  qed
  ultimately show ?case by blast
qed

lemma args_cstep_Hemb:
  assumes "length ss = length ts"
    and args: "\<forall>i<length ss. (ss ! i, ts ! i) \<in> (cstep (Hemb P))\<^sup>*"
  shows "(Fun f ss, Fun f ts) \<in> (cstep (Hemb P))\<^sup>*"
proof -
  have "ctxt_closed (cstep (Hemb P))" by (rule) auto
  from args_steps_imp_steps [OF this assms(1)] and assms(2)
    show ?thesis  by metis
qed

lemma Fun_cstep_Hemb:
  assumes "P\<^sup>=\<^sup>= g f" and "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb P))\<^sup>*) ts ss"
  shows "(Fun f ss, Fun g ts) \<in> (cstep (Hemb P))\<^sup>+"
proof -
  from list_hembeq_cstep_HembD [OF assms(2)] and args_cstep_Hemb
    obtain us where emb: "sublisteq us ss"
    and "(Fun g us, Fun g ts) \<in> (cstep (Hemb P))\<^sup>*" by force
  moreover have "(Fun f ss, Fun g us) \<in> (cstep (Hemb P))\<^sup>+"
  proof -
    from emb and `P\<^sup>=\<^sup>= g f` have "(Fun f ss, Fun g us) \<in> (cstep (Hemb P))"
      by (auto simp: Hemb_def Emb_def)
    thus ?thesis by auto
  qed
  ultimately show ?thesis by auto
qed

lemma csteps_ctxt:
  assumes "(s, t) \<in> (cstep R)\<^sup>+"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (cstep R)\<^sup>+"
  using assms by (induct) (auto intro: trancl_trans)

lemma hemb_imp_cstep_Hemb_trancl:
  assumes "hemb P s t"
  shows "(t, s) \<in> (cstep (Hemb P))\<^sup>+"
using assms
proof (induct)
  case (tree_hembeq_base t ts f)
  hence "(Fun f ts, t) \<in> Hemb P" by (auto simp: Emb_def Hemb_def)
  thus ?case by auto
next
  case (tree_hembeq_sublisteq f g ss ts)
  from this(2) show ?case
  proof (induct)
    case (list_hembeq_Nil ys)
    have "sublisteq [] ys" by auto
    with `P\<^sup>=\<^sup>= f g` have "(Fun g ys, Fun f []) \<in> Hemb P" by (auto simp: Hemb_def)
    thus ?case by auto 
  next
    case (list_hembeq_Cons xs ys y)
    have "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb P))\<^sup>*) xs ys"
      by (rule list_hembeq_mono [rule_format]) (insert list_hembeq_Cons, auto)
    hence "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb P))\<^sup>*) xs (y # ys)" by auto
    from Fun_cstep_Hemb [of P f g, OF `P\<^sup>=\<^sup>= f g` this] show ?case by simp
  next
    case (list_hembeq_Cons2 x y xs ys)
    have "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb P))\<^sup>*) xs ys"
      by (rule list_hembeq_mono [rule_format]) (insert list_hembeq_Cons2, auto)
    with list_hembeq_Cons2
      have "list_hembeq (\<lambda>s t. (t, s) \<in> (cstep (Hemb P))\<^sup>*) (x # xs) (y # ys)" by auto
    from Fun_cstep_Hemb [of P f g, OF `P\<^sup>=\<^sup>= f g` this] show ?case by simp
  qed
next
  case (tree_hembeq_trans s t u) thus ?case by auto
next
  case (tree_hembeq_ctxt s t f ss1 ss2)
  from csteps_ctxt [OF this(2), of "More f ss1 \<box> ss2"]
    show ?case by simp
qed

lemma cstep_Hemb_imp_hemb:
  assumes "(s, t) \<in> cstep (Hemb P)"
  shows "hemb P t s"
using assms
proof
  fix C l r
  assume "(l, r) \<in> Hemb P" and s: "s = C\<langle>l\<rangle>" and t: "t = C\<langle>r\<rangle>"
  thus ?thesis
  proof (unfold Hemb_def, elim UnE)
    assume "(l, r) \<in> Emb"
    then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" by (auto simp: Emb_def)
    hence "hemb P r l" unfolding l by auto
    from hemb_ctxt [OF this] show ?thesis by (simp add: s t)
  next
    assume "(l, r) \<in> {(Fun f ss, Fun g ts) | f ss g ts. P\<^sup>=\<^sup>= g f \<and> sublisteq ts ss}"
    then obtain f ss g ts where l: "l = Fun f ss" and r: "r = Fun g ts"
      and "sublisteq ts ss" and "P\<^sup>=\<^sup>= g f" by auto
    from `sublisteq ts ss` and s and t and l and r show ?thesis
    proof (induct arbitrary: l r)
      case (list_hembeq_Nil ss)
      have "list_hembeq (hemb P) [] ss" by auto
      with `P\<^sup>=\<^sup>= g f` have "hemb P r l" unfolding list_hembeq_Nil by auto
      from hemb_ctxt [OF this] show ?case unfolding list_hembeq_Nil by blast
    next
      case (list_hembeq_Cons ts ss u)
      moreover have "sublisteq ss (u # ss)" by auto
      ultimately have *: "sublisteq ts (u # ss)" by auto
      from term_tree.tree_hembeq_sublisteq [OF `P\<^sup>=\<^sup>= g f` this]
        have "hemb P r l"
        unfolding list_hembeq_Cons .
      from hemb_ctxt [OF this] show ?case unfolding list_hembeq_Cons by blast
    next
      case (list_hembeq_Cons2 u v ts ss)
      hence *: "sublisteq (u # ts) (v # ss)" by auto
      hence "hemb P r l" unfolding list_hembeq_Cons2 using `P\<^sup>=\<^sup>= g f` by auto
      from hemb_ctxt [OF this] show ?case unfolding list_hembeq_Cons2 by blast
    qed
  qed
qed

lemma cstep_Hemb_trancl_imp_hemb:
  assumes "(s, t) \<in> (cstep (Hemb P))\<^sup>+"
  shows "hemb P t s"
using assms
proof (induct s t)
  case (r_into_trancl s t)
  from this show ?case by (rule cstep_Hemb_imp_hemb)
next
  case (trancl_into_trancl s t u)
  from `(t, u) \<in> cstep (Hemb P)` have "hemb P u t" by (rule cstep_Hemb_imp_hemb)
  with trancl_into_trancl show ?case by blast
qed

lemma list_hembeq_eq_map:
  assumes "sublisteq ss ts"
  shows "sublisteq (map f ss) (map f ts)"
  using assms by (induct) auto

lemma Hemb_subst:
  assumes "(l, r) \<in> Hemb P"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> Hemb P"
    (is "(?l, ?r) \<in> ?R")
using assms [unfolded Hemb_def]
proof
  assume "(l, r) \<in> Emb"
  then obtain f ts where l: "l = Fun f ts" and "r \<in> set ts" by (auto simp: Emb_def)
  let ?ts = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  have "?l = Fun f ?ts" unfolding l by simp
  moreover have "?r \<in> set ?ts" using `r \<in> set ts` by auto
  ultimately show "(?l, ?r) \<in> ?R" by (auto simp: Hemb_def Emb_def)
next
  assume "(l, r) \<in> {(Fun f ss, Fun g ts) | f ss g ts. P\<^sup>=\<^sup>= g f \<and> sublisteq ts ss}"
  then obtain f g ss ts where "P\<^sup>=\<^sup>= g f" and l: "l = Fun f ss" and r: "r = Fun g ts"
    and emb: "sublisteq ts ss" by auto
  let ?ts = "map (\<lambda>t. t \<cdot> \<sigma>) ts"
  let ?ss = "map (\<lambda>t. t \<cdot> \<sigma>) ss"
  let ?l = "l \<cdot> \<sigma>"
  let ?r = "r \<cdot> \<sigma>"
  from list_hembeq_eq_map [OF emb] have "sublisteq ?ts ?ss" by auto
  with `P\<^sup>=\<^sup>= g f` show "(?l, ?r) \<in> ?R"
    unfolding Hemb_def l r by auto
qed

lemma cstep_Hemb_eq_rstep_Hemb:
  "cstep (Hemb P) = rstep (Hemb P)" (is "?L = ?R")
proof
  show "?R \<subseteq> ?L"
  proof
    fix s t
    assume "(s, t) \<in> ?R"
    thus "(s, t) \<in> ?L"
    proof
      fix l r C \<sigma>
      assume "(l, r) \<in> Hemb P" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      from Hemb_subst [OF this(1)] have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> cstep (Hemb P)" by auto
      from cstep.step [OF this] show ?thesis unfolding s t by blast
    qed
  qed
next
  show "?L \<subseteq> ?R"
  proof
    fix s t
    assume "(s, t) \<in> ?L"
    thus "(s, t) \<in> ?R"
      by (induct s t) auto
  qed
qed

lemma hemb_eq_rstep_Hemb_trancl:
  "hemb P s t \<longleftrightarrow> (t, s) \<in> (rstep (Hemb P))\<^sup>+"
  unfolding cstep_Hemb_eq_rstep_Hemb [symmetric]
  using hemb_imp_cstep_Hemb_trancl
    and cstep_Hemb_trancl_imp_hemb by blast

lemma wqo_on_hemb:
  assumes "wqo_on P A"
  shows "wqo_on (hemb P) (gterms A)"
  using term_tree.wqo_on_trees [OF assms] by simp

lemmas wfp_on_hemb = wqo_on_hemb [THEN wqo_on_imp_wfp_on]

end

